#include "LibSip_SIPTypesAndValues.hh"
#include "LibSip_Common.hh"

#include "sip_codec_headers.hh"
#include "loggers.hh"

#include "converter.hh"

int sip_codec_headers::encode_sip_url(const LibSip__SIPTypesAndValues::SipUrl& p_sip_uri, osip_uri_t** p_uri)
{ // TODO To be renamed into encode_uri
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_sip_uri: ", p_sip_uri);

  std::string host;
  std::string port;
  osip_uri_t *uri = nullptr;
  ::osip_uri_init(&uri);
  const LibSip__SIPTypesAndValues::UriComponents& components = p_sip_uri.components();
  if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_sip)) {
    ::osip_uri_set_scheme(uri, (char*)static_cast<const char*>(p_sip_uri.scheme()));
    const LibSip__SIPTypesAndValues::SipUriComponents& s = p_sip_uri.components().sip();
    if (s.userInfo().is_present()) {
      const LibSip__SIPTypesAndValues::UserInfo& u = static_cast<const LibSip__SIPTypesAndValues::UserInfo>(s.userInfo());
      ::osip_uri_set_username(uri, (char*)static_cast<const char*>(static_cast<CHARSTRING>(u.userOrTelephoneSubscriber())));
      if (u.password().is_present()) {
        const CHARSTRING& c = static_cast<CHARSTRING>(u.password());
        ::osip_uri_set_password(uri, (char*)static_cast<const char*>(c));
      }
    }
    encode_host_port(s.hostPort(), host, port);
    if (!host.empty()) {
      ::osip_uri_set_host(uri, (char*)host.c_str());
    }
    if (!port.empty()) {
     ::osip_uri_set_port(uri, (char*)port.c_str());
    }
  } else if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_tel)) {
    const LibSip__SIPTypesAndValues::TelUriComponents& t = p_sip_uri.components().tel();
    loggers::get_instance().error("sip_codec_headers::encode_sip_uri: Unsupported LibSip__SIPTypesAndValues::UriComponents::ALT_tel");
  } else if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_urn)) {
    const LibSip__SIPTypesAndValues::UrnUriComponents& u = p_sip_uri.components().urn();
    loggers::get_instance().log("sip_codec_headers::encode_sip_uri: Decode Urn");
    std::string str(static_cast<const char*>(p_sip_uri.scheme()));
    str += ":";
    str += static_cast<const char*>(u.namespaceId());
    str += ":";
    str += static_cast<const char*>(u.namespaceSpecificString());
    ::osip_uri_parse(uri, str.c_str());
  } else if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_other)) {
    loggers::get_instance().error("sip_codec_headers::encode_sip_uri: Unsupported LibSip__SIPTypesAndValues::UriComponents::ALT_other");
  } // else, noting to do
  
  if (uri != nullptr) {
    ::osip_uri_clone(uri, p_uri);
  } else {
    *p_uri = nullptr;
  }
  
  char* buffer = nullptr;
  size_t length = 0;
  int result = ::osip_uri_to_str(*p_uri, &buffer);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_headers::encode_sip_uri: Failed to encode data structures");
  } else {
    loggers::get_instance().log("sip_codec_headers::encode_sip_uri: URI:%s", buffer);
  }
  
  return 0;
} // End of method encode_sip_url

void sip_codec_headers::encode_host_port(const LibSip__SIPTypesAndValues::HostPort& p_host_port, std::string& p_host, std::string& p_port)
{
  loggers::get_instance().log(">>> sip_codec_headers::encode_host_port");
  
  if (p_host_port.host().is_present()) {
    p_host.assign(static_cast<const char*>(static_cast<CHARSTRING>(p_host_port.host())));
  } else {
    p_host.clear();
  }
  loggers::get_instance().log("sip_codec_headers::encode_host_port: host:'%s'", p_host.c_str());
  if (p_host_port.portField().is_present()) {
    p_port.assign(std::to_string(static_cast<int>(static_cast<INTEGER>(p_host_port.portField()))));
  } else {
    p_port.clear();
  }
  loggers::get_instance().log("sip_codec_headers::encode_port_port: port:'%s'", p_port.c_str());
} // End of method encode_host_port

int sip_codec_headers::encode_accept_header(const LibSip__SIPTypesAndValues::Accept& p_accept, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_accept_header");
  
  if (!p_accept.acceptArgs().is_present()) {
    return 0;
  }
  
  const LibSip__SIPTypesAndValues::AcceptBody__List& al = static_cast<const LibSip__SIPTypesAndValues::AcceptBody__List&>(*p_accept.acceptArgs().get_opt_value());
  if (al.lengthof() == 0) {
    return 0;
  }
  std::string accepts;
  for (int i = 0; i < al.lengthof(); i++) {
    const LibSip__SIPTypesAndValues::AcceptBody& b = al[i];
    loggers::get_instance().log_msg("sip_codec_headers::encode_accept_header: b: ", b);
    osip_accept_t *accept;
    accept_init(&accept);
    // Split it using '/' as separator
    std::vector<std::string> output = converter::get_instance().split(static_cast<const char*>(b.mediaRange()), "/");
    accept->type = ::strdup(output[0].c_str());
    if (output.size() > 1) {
      accept->subtype = ::strdup(output[1].c_str());
    }
    if (b.acceptParam().is_present()) {
      encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(b.acceptParam()), &(accept->gen_params));
    }
    char *buff;
    ::osip_accept_to_str(accept, &buff);
    accepts += (const char*)buff;
    accepts += ",";
    osip_free(buff);
    ::osip_accept_free(accept);
    loggers::get_instance().log("sip_codec_headers::encode_accept_header: accepts: %s", accepts.c_str());
  } // End of 'for' statement
  accepts.resize(accepts.length() - 1); // Remove the last ','
  ::osip_message_set_accept(*p_sip_message, ::strdup(accepts.c_str()));
  
  return 0;
} // End of method encode_accept_header

int sip_codec_headers::encode_accept_contact_header(const LibSip__SIPTypesAndValues::AcceptContact& p_accept_contact, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_accept_contact_header");

  const LibSip__SIPTypesAndValues::AcRcValue__List& al = p_accept_contact.acValues();
  if (al.lengthof() == 0) {
    return 0;
  }

  std::string accepts;
  for (int i = 0; i < al.lengthof(); i++) {
    const LibSip__SIPTypesAndValues::AcRcValue& a = al[i];
    loggers::get_instance().log_msg("sip_codec_headers::encode_accept_contact_header: a: ", a);
    accepts += static_cast<const char*>(a.wildcard());
    if (a.acRcParams().is_present()) {
      const LibSip__Common::SemicolonParam__List& l = static_cast<const LibSip__Common::SemicolonParam__List&>(*a.acRcParams().get_opt_value());
      for (int j = 0; j < l.lengthof(); j++) {
        accepts += ";";
        const LibSip__Common::GenericParam& p = l[j];
        accepts += static_cast<const char*>(p.id());
        if (p.paramValue().is_present()) {
          const LibSip__Common::GenValue& g = static_cast<const LibSip__Common::GenValue&>(*p.paramValue().get_opt_value());;
          accepts += "=";
          if (g.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
            accepts += static_cast<const char*>(g.tokenOrHost());
          } else {
            accepts += static_cast<const char*>(g.quotedString());
          }
        }
      } // End of 'for' statement
    }
    accepts += ",";
    loggers::get_instance().log("sip_codec_headers::encode_accept_header: accepts: %s", accepts.c_str());
  } // End of 'for' statement
  accepts.resize(accepts.length() - 1); // Remove the last ','
  ::osip_message_set_header(*p_sip_message, (const char *)"Accept-Contact", ::strdup(accepts.c_str()));

  loggers::get_instance().log("<<< sip_codec_headers::encode_accept_contact_header");
  return 0;
} // End of method encode_accept_contact_header

int sip_codec_headers::encode_allow_header(const OPTIONAL<LibSip__SIPTypesAndValues::Allow>& p_allow, osip_message_t** p_sip_message)
{
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_allow_header", p_allow);

  const LibSip__SIPTypesAndValues::Allow& allow = static_cast<const LibSip__SIPTypesAndValues::Allow&>(*p_allow.get_opt_value());

  if (!allow.methods().is_present()) {
    return 0;
  }
  
  const LibSip__SIPTypesAndValues::Method__List& m = static_cast<const LibSip__SIPTypesAndValues::Method__List&>(*allow.methods().get_opt_value());
  if (m.lengthof() == 0) {
    return 0;
  }
  
  std::string str(static_cast<const char*>(m[0]));
  if (m.lengthof() > 1) {
    int i = 1;
    do {
      str += ",";
      str += static_cast<const char*>(m[i++]);
    } while (i < m.lengthof());
  }
  ::osip_message_set_header((osip_message_t *)*p_sip_message,(const char *)"Allow", str.c_str());
  
  loggers::get_instance().log("<<< sip_codec_headers::encode_allow_header");
  return 0;
} // End of method encode_allow_header

int sip_codec_headers::encode_authorization_header(const LibSip__SIPTypesAndValues::Authorization& p_authorization, osip_authorization_t** p_authorization_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::encode_authorization_header");
  
  ::osip_authorization_init(p_authorization_header);

  const LibSip__SIPTypesAndValues::CredentialsList& l = p_authorization.body();
  int i = 0;
  do {
    loggers::get_instance().log("sip_codec_headers::encode_authorization_header: Processing item #%d", i);
    const LibSip__SIPTypesAndValues::Credentials& c = l[i++];
    if (c.ischosen(LibSip__SIPTypesAndValues::Credentials::ALT_digestResponse)) {
      bool processed = true;
      const LibSip__Common::CommaParam__List& p = c.digestResponse();
      if (p.lengthof() > 0) {
        loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: Processing param ", p);
        int j = 0;
        do {
          const LibSip__Common::GenericParam& g = p[j++];
          std::string str(static_cast<const char*>(g.id()));
          if (str.compare("realm") == 0) {
            loggers::get_instance().log("sip_codec_headers::encode_authorization_header: realm found");
            if (g.paramValue().is_present()) {
              const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
              loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: GenValue: ", v);
              if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
                loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: tokenOrHost: ", v.tokenOrHost());
                ::osip_authorization_set_realm(*p_authorization_header, (char*)::strdup(static_cast<const char*>(v.tokenOrHost())));
              } else {
                loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented (3)");
              }
            } else {
              loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented(4)");
            }
          } else if (str.compare("username") == 0) {
            loggers::get_instance().log("sip_codec_headers::encode_authorization_header: username found");
            if (g.paramValue().is_present()) {
              const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
              loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: GenValue: ", v);
              if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
                loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: tokenOrHost: ", v.tokenOrHost());
                ::osip_authorization_set_username(*p_authorization_header, (char*)::strdup(static_cast<const char*>(v.tokenOrHost())));
              } else {
                loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented (5)");
              }
            } else {
              loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented(6)");
            }
          } else if (str.compare("uri") == 0) {
            loggers::get_instance().log("sip_codec_headers::encode_authorization_header: uri found");
            if (g.paramValue().is_present()) {
              const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
              loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: GenValue: ", v);
              if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
                loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: tokenOrHost: ", v.tokenOrHost());
                ::osip_authorization_set_uri(*p_authorization_header, (char*)::strdup(static_cast<const char*>(v.tokenOrHost())));
              } else {
                loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented (7)");
              }
            } else {
              loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented(7)");
            }
          } else if (str.compare("nonce") == 0) {
            loggers::get_instance().log("sip_codec_headers::encode_authorization_header: nonce found");
            if (g.paramValue().is_present()) {
              const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
              loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: GenValue: ", v);
              if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
                loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: tokenOrHost: ", v.tokenOrHost());
                ::osip_authorization_set_nonce(*p_authorization_header, (char*)::strdup(static_cast<const char*>(v.tokenOrHost())));
              } else {
                loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented (9)");
              }
            } else {
              loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented(10)");
            }
          } else if (str.compare("response") == 0) {
            loggers::get_instance().log("sip_codec_headers::encode_authorization_header: response found");
            if (g.paramValue().is_present()) {
              const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
              loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: GenValue: ", v);
              if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
                loggers::get_instance().log_msg("sip_codec_headers::encode_authorization_header: tokenOrHost: ", v.tokenOrHost());
                ::osip_authorization_set_response(*p_authorization_header, (char*)::strdup(static_cast<const char*>(v.tokenOrHost())));
              } else {
                loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented (11)");
              }
            } else {
              loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented(12)");
            }
          } else {
            loggers::get_instance().log("sip_codec_headers::encode_authorization_header: Set processed to false for %s", str.c_str());
            processed = false;
          }
        } while (j < p.lengthof());
      }
      loggers::get_instance().log("sip_codec_headers::encode_authorization_header: processed: %x", processed);
      if (processed) {
        ::osip_authorization_set_auth_type(*p_authorization_header, (char*)::strdup("Digest"));
      }
    } else {
      const LibSip__SIPTypesAndValues::OtherAuth& o = c.otherResponse();
      loggers::get_instance().error("sip_codec_headers::encode_authorization_header: Not implemented");
    }
  } while (i < l.lengthof());
  
  return 0;
} // End of method encode_authorization_header

int sip_codec_headers::encode_call_id_header(const LibSip__SIPTypesAndValues::CallId& p_call_id, osip_call_id_t** p_call_id_header) {
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_call_id_header", p_call_id.callid());
  
  ::osip_call_id_init(p_call_id_header);
  ::osip_call_id_parse(*p_call_id_header, static_cast<const char*>(p_call_id.callid()));
  
  loggers::get_instance().log("<<< sip_codec_headers::encode_call_id_header");
  return 0;
} // End of method encode_call_id_header

int sip_codec_headers::encode_contact_header(const LibSip__SIPTypesAndValues::Contact& p_contact, osip_contact_t** p_contact_header) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_contact_header");

  ::osip_contact_init(p_contact_header);
  const LibSip__SIPTypesAndValues::ContactBody& body = p_contact.contactBody();
  if (body.ischosen(LibSip__SIPTypesAndValues::ContactBody::ALT_wildcard)) {
    const CHARSTRING& wildcard = body.wildcard();
    loggers::get_instance().warning("sip_codec_headers::encode_contact_header: wildcard not implemented yet");
    *p_contact_header = nullptr;
    return -1;
  } else if (body.ischosen(LibSip__SIPTypesAndValues::ContactBody::ALT_contactAddresses)) {
    const LibSip__SIPTypesAndValues::ContactAddress__List& l = body.contactAddresses();
    // Encode AddressField
    for (int i = 0; i < l.size_of(); i++) {
      osip_uri_t* uri = nullptr;
      const LibSip__SIPTypesAndValues::ContactAddress c = l[i];
      if (c.addressField().ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_nameAddr)) {
        const LibSip__SIPTypesAndValues::NameAddr& addr = c.addressField().nameAddr();
        if (encode_sip_url(addr.addrSpec(), &uri) == -1) {
          loggers::get_instance().warning("sip_codec_headers::encode_contact_header: Failed to encode SipUrl");
          ::osip_contact_free(*p_contact_header);
          *p_contact_header = nullptr;
          return -1;
        }
        ::osip_contact_set_url(*p_contact_header, uri);
        if (addr.displayName().is_present()) {
          const LibSip__SIPTypesAndValues::DisplayName& n = static_cast<const LibSip__SIPTypesAndValues::DisplayName&>(addr.displayName());
          if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_token)) {
            ::osip_contact_set_displayname(*p_contact_header, (char*)::strdup(static_cast<const char*>(static_cast<CHARSTRING>(n.token()))));
          } else if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_quotedString)) {
            ::osip_contact_set_displayname(*p_contact_header, (char*)::strdup(static_cast<const char*>(static_cast<CHARSTRING>(n.quotedString()))));
          } else {
            loggers::get_instance().warning("sip_codec_headers::encode_contact_header: Failed to encode DisplayName");
            ::osip_contact_free(*p_contact_header);
            *p_contact_header = nullptr;
            return -1;
          }
        }
      } else if (c.addressField().ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_addrSpecUnion)) {
        if (encode_sip_url(c.addressField().addrSpecUnion(), &uri) == -1) {
          loggers::get_instance().warning("sip_codec_headers::encode_contact_header: Failed to encode SipUrl");
          ::osip_contact_free(*p_contact_header);
          *p_contact_header = nullptr;
          return -1;
        }
        ::osip_contact_set_url(*p_contact_header, uri);
      } else {
        loggers::get_instance().warning("sip_codec_headers::encode_contact_header: Failed to encode Addr__Union");
        ::osip_contact_free(*p_contact_header);
        *p_contact_header = nullptr;
        return -1;
      }
      // Encode contactParams
      if (c.contactParams().is_present()) {
        encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(c.contactParams()), &(*p_contact_header)->gen_params);
      }

    } // End of 'for' statement
  } else {
    return -1;
  }
  
  loggers::get_instance().log("<<< sip_codec_headers::encode_contact_header");
  return 0;
} // End of method encode_contact_header

int sip_codec_headers::encode_content_length_header(const LibSip__SIPTypesAndValues::ContentLength& p_content_length, osip_content_length_t** p_content_length_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::encode_content_length_header");

  ::osip_content_length_init(p_content_length_header);
  (*p_content_length_header)->value = (char*)::strdup(std::to_string(static_cast<const int>(p_content_length.len())).c_str());
  
  loggers::get_instance().log("<<< sip_codec_headers::encode_content_length_header");
  return 0;
} // End of method encode_content_length_header

int sip_codec_headers::encode_content_type_header(const LibSip__SIPTypesAndValues::ContentType& p_content_type, osip_content_type_t** p_content_type_header)
{
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_content_type_header: ", p_content_type);

  ::osip_content_type_init(p_content_type_header);

  // Split type/subtype using '/' as separator
  std::vector<std::string> output = converter::get_instance().split(static_cast<const char*>(p_content_type.mTypeSubtype()), "/");
  (*p_content_type_header)->type = ::strdup(output[0].c_str());
  if (output.size() > 1) {
    (*p_content_type_header)->subtype = ::strdup(output[1].c_str());
  }
  
  if (p_content_type.mParams().is_present()) {
    encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(p_content_type.mParams()), &((*p_content_type_header)->gen_params));
  }
  
  loggers::get_instance().log("<<< sip_codec_headers::encode_content_type_header");
  return 0;
} // End of method encode_content_type_header

int sip_codec_headers::encode_c_seq_header(const LibSip__SIPTypesAndValues::CSeq& p_c_seq, osip_cseq_t** p_c_seq_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::encode_c_seq_header");

  ::osip_cseq_init(p_c_seq_header);
  ::osip_cseq_set_number(*p_c_seq_header, (char*)::strdup(std::to_string(static_cast<const int>(p_c_seq.seqNumber())).c_str()));
  ::osip_cseq_set_method(*p_c_seq_header, (char*)::strdup(static_cast<const char*>(p_c_seq.method())));
  
  return 0;
} // End of method encode_c_seq_header

int sip_codec_headers::encode_from_header(const LibSip__SIPTypesAndValues::From& p_from, osip_from_t** p_from_header) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_from_header");
  
  ::osip_from_init(p_from_header);
  // Endode addressField
  osip_uri_t* uri = nullptr;
  const LibSip__SIPTypesAndValues::Addr__Union& a = p_from.addressField();
  if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_nameAddr)) {
    const LibSip__SIPTypesAndValues::NameAddr& addr = a.nameAddr();
    if (encode_sip_url(addr.addrSpec(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_headers::encode_from_header: Failed to encode SipUrl");
      ::osip_from_free(*p_from_header);
      *p_from_header = nullptr;
      return -1;
    }
    ::osip_from_set_url(*p_from_header, uri);
    if (addr.displayName().is_present()) {
      const LibSip__SIPTypesAndValues::DisplayName& n = static_cast<const LibSip__SIPTypesAndValues::DisplayName&>(addr.displayName());
      if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_token)) {
        ::osip_from_set_displayname(*p_from_header, (char*)::strdup(static_cast<const char*>(static_cast<CHARSTRING>(n.token()))));
      } else if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_quotedString)) {
        ::osip_from_set_displayname(*p_from_header, (char*)::strdup(static_cast<const char*>(static_cast<CHARSTRING>(n.quotedString()))));
      } else {
        loggers::get_instance().warning("sip_codec_headers::encode_from_header: Failed to encode DisplayName");
        ::osip_from_free(*p_from_header);
        *p_from_header = nullptr;
        return -1;
      }
    }
  } else if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_addrSpecUnion)) {
    if (encode_sip_url(a.addrSpecUnion(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_headers::encode_from_header: Failed to encode SipUrl");
      ::osip_from_free(*p_from_header);
      *p_from_header = nullptr;
      return -1;
    }
    ::osip_from_set_url(*p_from_header, uri);
  } else {
    loggers::get_instance().warning("sip_codec_headers::encode_from_header: Failed to encode Addr__Union");
    ::osip_from_free(*p_from_header);
    *p_from_header = nullptr;
    return -1;
  }
  // Encode fromParams
  if (p_from.fromParams().is_present()) {
    encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(p_from.fromParams()), &(*p_from_header)->gen_params);
  }
  
  return 0;
} // End of method encode_from_header

int sip_codec_headers::encode_max_forwards_header(const OPTIONAL<LibSip__SIPTypesAndValues::MaxForwards>& p_max_forwards, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_max_forwards_header");

  if (!p_max_forwards.is_present()) {
    return 0;
  }
  
  const LibSip__SIPTypesAndValues::MaxForwards& max_forwards = static_cast<const LibSip__SIPTypesAndValues::MaxForwards&>(*p_max_forwards.get_opt_value());
  osip_message_set_max_forwards(*p_sip_message, std::to_string(static_cast<int>(max_forwards.forwards())).c_str());
  
  return 0;
}

int sip_codec_headers::encode_min_se_header(const OPTIONAL<LibSip__SIPTypesAndValues::MinSE>& p_min_se, osip_message_t** p_sip_message) {
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_min_se_header", p_min_se);

  const LibSip__SIPTypesAndValues::MinSE& min_se = static_cast<const LibSip__SIPTypesAndValues::MinSE&>(*p_min_se.get_opt_value());
  // deltaSec
  std::string value(static_cast<const char*>(min_se.deltaSec()));
  // seParam
  const OPTIONAL<LibSip__Common::SemicolonParam__List>& se_param = min_se.minSeParam(); // TODO Create a methid to fill an std::string with SemicolonParam__List
  if (se_param.is_present()) {
    const LibSip__Common::SemicolonParam__List& l = static_cast<const LibSip__Common::SemicolonParam__List&>(*se_param.get_opt_value());
    int i = 0;
    do {
      value += ";";
      const LibSip__Common::GenericParam& g = l[i];
      value+= static_cast<const char*>(g.id());
      if (g.paramValue().is_present()) {
        value += "=";
        const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
        if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
          value += static_cast<const char*>(v.tokenOrHost());
        } else {
          value += static_cast<const char*>(v.quotedString());
        }
      }
    } while (i < l.lengthof());
  }
  ::osip_message_set_header((osip_message_t *)*p_sip_message,(const char *)"Min-SE", value.c_str());
  
  return 0;
}

int sip_codec_headers::encode_session_expires_header(const OPTIONAL<LibSip__SIPTypesAndValues::SessionExpires>& p_session_expires, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_session_expires_header");

  if (!p_session_expires.is_present()) {
    return 0;
  }
  
  const LibSip__SIPTypesAndValues::SessionExpires& session_expires = static_cast<const LibSip__SIPTypesAndValues::SessionExpires&>(*p_session_expires.get_opt_value());
  // deltaSec
  std::string value(static_cast<const char*>(session_expires.deltaSec()));
  // seParam
  const OPTIONAL<LibSip__Common::SemicolonParam__List>& se_param = session_expires.seParam(); // TODO Create a methid to fill an std::string with SemicolonParam__List
  if (se_param.is_present()) {
    const LibSip__Common::SemicolonParam__List& l = static_cast<const LibSip__Common::SemicolonParam__List&>(*se_param.get_opt_value());
    int i = 0;
    do {
      value += ";";
      const LibSip__Common::GenericParam& g = l[i];
      value+= static_cast<const char*>(g.id());
      if (g.paramValue().is_present()) {
        value += "=";
        const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*g.paramValue().get_opt_value());
        if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
          value += static_cast<const char*>(v.tokenOrHost());
        } else {
          value += static_cast<const char*>(v.quotedString());
        }
      }
    } while (i < l.lengthof());
  }
  osip_message_set_expires(*p_sip_message, value.c_str());
  
  return 0;
}

int sip_codec_headers::encode_supported_header(const OPTIONAL<LibSip__SIPTypesAndValues::Supported>& p_supported, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_supported_header");

  if (!p_supported.is_present()) {
    return 0;
  }
  
  const LibSip__SIPTypesAndValues::Supported& supported = static_cast<const LibSip__SIPTypesAndValues::Supported&>(*p_supported.get_opt_value());
  const OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>& tags = supported.optionsTags();
  if (!tags.is_present()) {
    return 0;
  }
  if (encode_option_tag_list(static_cast<const LibSip__SIPTypesAndValues::OptionTag__List&>(*tags.get_opt_value()), p_sip_message) == -1) {
    return -1;
  }
  
  return 0;
}

int sip_codec_headers::encode_p_access_network_info_header(const OPTIONAL<LibSip__SIPTypesAndValues::PAccessNetworkInfo>& p_access_network_info, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_p_access_network_info_header");

  if (!p_access_network_info.is_present()) {
    return 0;
  }

  // pAccessNetworkInfo := { fieldName := P_ACCESS_NETWORK_INFO_E (65), accessType := "IEEE-802.11a", genericParams := { { id := "extension-access-info", paramValue := { tokenOrHost := "192.1.1.20" } } } }
  // Encode AccessType
  // TODO Try to get the P-Access-Network-Info header
  
  /*osip_header_t* p_access_network_info_;
  ::osip_header_init(&p_access_network_info_);
  osip_header_set_name (p_access_network_info_, "P-Access-Network-Info";
  // Encode generic params
  const LibSip__SIPTypesAndValues::PAccessNetworkInfo& pa = static_cast<const LibSip__SIPTypesAndValues::PAccessNetworkInfo&>(*p_access_network_info.get_opt_value());
  if (pa.genericParams().is_present()) {
    osip__t* params;
    ::osip_list_init(&params);
    encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(pa.genericParams()), &params);
    
    osip_header_set_value(p_access_network_info_, value);
    //p_access_network_info_->hvalue = 
  }
  
  osip_message_set_header(p_sip_message, "P-Access-Network-Info", p_access_network_info_);
  ::osip_header_free(p_access_network_info_);*/
  
  return 0;
}

int sip_codec_headers::encode_to_header(const LibSip__SIPTypesAndValues::To& p_to, osip_to_t** p_to_header) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_to_header");
  
  ::osip_to_init(p_to_header);
  const LibSip__SIPTypesAndValues::Addr__Union& a = p_to.addressField();
  if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_nameAddr)) {
    const LibSip__SIPTypesAndValues::NameAddr& addr = a.nameAddr();
    osip_uri_t *uri;
    if (encode_sip_url(addr.addrSpec(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_headers::encode_to_header: Failed to encode SipUrl");
      ::osip_to_free(*p_to_header);
      *p_to_header = nullptr;
      return -1;
    }
    ::osip_to_set_url(*p_to_header, uri);
    //::osip_uri_free(uri);
    if (addr.displayName().is_present()) {
      const LibSip__SIPTypesAndValues::DisplayName& n = static_cast<const LibSip__SIPTypesAndValues::DisplayName&>(addr.displayName());
      if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_token)) {
        ::osip_to_set_displayname(*p_to_header, (char*)::strdup(static_cast<const char*>(static_cast<CHARSTRING>(n.token()))));
      } else if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_quotedString)) {
        ::osip_to_set_displayname(*p_to_header, (char*)::strdup(static_cast<const char*>(static_cast<CHARSTRING>(n.quotedString()))));
      } else {
        loggers::get_instance().warning("sip_codec_headers::encode_to_header: Failed to encode DisplayName");
        ::osip_to_free(*p_to_header);
        *p_to_header = nullptr;
        return -1;
      }
    }
  } else if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_addrSpecUnion)) {
    osip_uri_t *uri;
    if (encode_sip_url(a.addrSpecUnion(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_headers::encode_to_header: Failed to encode SipUrl");
      ::osip_to_free(*p_to_header);
      *p_to_header = nullptr;
      return -1;
    }
    ::osip_to_set_url(*p_to_header, uri);
  } else {
    loggers::get_instance().warning("sip_codec_headers::encode_to_header: Failed to encode Addr__Union");
    ::osip_to_free(*p_to_header);
    *p_to_header = nullptr;
    return -1;
  }
  if (p_to.toParams().is_present()) {
    encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(p_to.toParams()), &(*p_to_header)->gen_params);
  }
  
  return 0;
} // End of method encode_to_header

int sip_codec_headers::encode_via_header(const LibSip__SIPTypesAndValues::Via& p_via, osip_via_t** p_via_header) {
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_via_header: ", p_via);
  
  ::osip_via_init(p_via_header);
  const LibSip__SIPTypesAndValues::ViaBody__List& l = p_via.viaBody();
  for (int i = 0; i < l.size_of(); i++) {
    const LibSip__SIPTypesAndValues::ViaBody v = l[i];
    loggers::get_instance().log_msg("sip_codec_headers::encode_via_header: Processing ", v);

    ::via_set_protocol(*p_via_header, (char*)::strdup(static_cast<const char *>(v.sentProtocol().protocolName())));
    ::via_set_version(*p_via_header, (char*)::strdup(static_cast<const char *>(v.sentProtocol().protocolVersion())));
    ::via_set_comment(*p_via_header, (char*)::strdup(static_cast<const char *>(v.sentProtocol().transport())));
    std::string host;
    std::string port;
    encode_host_port(v.sentBy(), host, port);
    if (!host.empty()) {
      ::via_set_host(*p_via_header, (char*)::strdup(host.c_str()));
    }
    if (!port.empty()) {
      ::via_set_port (*p_via_header, (char*)::strdup(port.c_str()));
    }
    
    if (v.viaParams().is_present()) {
      encode_semi_colon_params(static_cast<const LibSip__Common::SemicolonParam__List>(v.viaParams()), &(*p_via_header)->via_params);
    }
  } // End of 'for' statement
  
  return 0;
} // End of method encode_via_header
 
int sip_codec_headers::encode_semi_colon_params(const LibSip__Common::SemicolonParam__List& p_list, osip_list_t* p_sip_list) {
  loggers::get_instance().log_msg(">>> sip_codec_headers::encode_semi_colon_params: ", p_list);

  if (p_list.size_of() != 0) {
    for (int i = 0; i < p_list.size_of(); i++) {
      const LibSip__Common::GenericParam& param = static_cast<const LibSip__Common::GenericParam&>(*p_list.get_at(i));
      loggers::get_instance().log_msg("sip_codec_headers::encode_semi_colon_params: param: ", param);
      if (param.paramValue().is_present()) {
        const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*param.paramValue().get_opt_value());
        loggers::get_instance().log_msg("sip_codec_headers::encode_semi_colon_params: v: ", v);
        if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
          loggers::get_instance().log_msg("sip_codec_headers::encode_semi_colon_params: tokenOrHost: ", v.tokenOrHost());
          ::osip_generic_param_add(p_sip_list, (char*)::strdup(static_cast<const char*>(param.id())), (char*)::strdup(static_cast<const char*>(v.tokenOrHost())));
        } else {
          loggers::get_instance().log_msg("sip_codec_headers::encode_semi_colon_params: quotedString: ", v.quotedString());
          ::osip_generic_param_add(p_sip_list, (char*)static_cast<const char*>(param.id()), (char*)static_cast<const char*>(v.quotedString()));
        }
      }
    } // End of 'for' statement
  } else {
    // Nothing to do
  }
  
  return 0;
}

void sip_codec_headers::decode_semi_colon_params(const osip_list_t& p_sip_list, OPTIONAL<LibSip__Common::SemicolonParam__List>& p_list) {
  loggers::get_instance().log(">>> sip_codec_headers::decode_semi_colon_params");

  if (::osip_list_size(&p_sip_list) != 0) {
    LibSip__Common::SemicolonParam__List l;
    for (int i = 0; i < ::osip_list_size(&p_sip_list); i++) {
      osip_generic_param_t* p = (osip_generic_param_t*)::osip_list_get(&p_sip_list, i);
      loggers::get_instance().log("sip_codec_headers::decode_semi_colon_params: After cast: %p", p);
      CHARSTRING name(p->gname);
      if (p->gvalue == nullptr) {
        l[i] = LibSip__Common::GenericParam(name, OPTIONAL<LibSip__Common::GenValue>(OMIT_VALUE));
      } else {
        LibSip__Common::GenValue g;
        g.tokenOrHost() = CHARSTRING(p->gvalue);
        l[i] = LibSip__Common::GenericParam(name, g);
      }
    } // End of 'for' statement
    p_list = OPTIONAL<LibSip__Common::SemicolonParam__List>(l);
  } else {
    p_list.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_semi_colon_params: ", p_list);
} // End of method decode_semi_colon_params

void sip_codec_headers::decode_semi_colon_params(const std::string& p_params, std::string& p_first_param, LibSip__Common::SemicolonParam__List& p_others_params) {
  loggers::get_instance().log(">>> sip_codec_headers::decode_semi_colon_params: %s", p_params.c_str());
  
  std::size_t idx = p_params.find(";");
  if (idx == std::string::npos) {
    p_first_param = p_params;
  } else {
    std::vector<std::string> output = converter::get_instance().split(p_params, ";");
    p_first_param = output[0];
    LibSip__Common::SemicolonParam__List l;
    for (unsigned int i = 1; i < output.size(); i++) {
      std::vector<std::string> o = converter::get_instance().split(output[i], "=");
      LibSip__Common::GenValue g;
      g.tokenOrHost() = CHARSTRING(o[1].c_str());
      p_others_params[i - 1] = LibSip__Common::GenericParam(CHARSTRING(o[0].c_str()), g);
    } // End of 'for' statement
  }
  loggers::get_instance().log("<<< sip_codec_headers::decode_semi_colon_params: %s", p_first_param.c_str());
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_semi_colon_params: ", p_others_params);
}

int sip_codec_headers::encode_ampersand_params(const LibSip__Common::AmpersandParam__List& p_list, osip_list_t** p_sip_list) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_ampersand_params");

  if (p_list.size_of() != 0) {
    ::osip_list_init(*p_sip_list);
    for (int i = 0; i < p_list.size_of(); i++) {
      const LibSip__Common::GenericParam& param = static_cast<const LibSip__Common::GenericParam&>(*p_list.get_at(i));
      if (param.paramValue().is_present()) {
        const LibSip__Common::GenValue& v = static_cast<const LibSip__Common::GenValue&>(*param.paramValue().get_opt_value());
        if (v.ischosen(LibSip__Common::GenValue::ALT_tokenOrHost)) {
          ::osip_generic_param_add(*p_sip_list, (char*)static_cast<const char*>(param.id()), (char*)static_cast<const char*>(v.tokenOrHost()));
        } else {
          ::osip_generic_param_add(*p_sip_list, (char*)static_cast<const char*>(param.id()), (char*)static_cast<const char*>(v.quotedString()));
        }
      }
    } // End of 'for' statement
  } else {
    *p_sip_list = nullptr;
  }
  
  return 0;
}

void sip_codec_headers::decode_ampersand_params(const osip_list_t& p_sip_list, OPTIONAL<LibSip__Common::AmpersandParam__List>& p_list) {
  loggers::get_instance().log(">>> sip_codec_headers::decode_ampersand_params");

  if (::osip_list_size(&p_sip_list) != 0) {
    LibSip__Common::AmpersandParam__List l;
    for (int i = 0; i < ::osip_list_size(&p_sip_list); i++) {
      osip_generic_param_t* p = (osip_generic_param_t*)::osip_list_get(&p_sip_list, i);
      loggers::get_instance().log("sip_codec_headers::decode_ampersand_params: After cast: %p", p);
      CHARSTRING name(p->gname);
      if (p->gvalue == nullptr) {
        l[i] = LibSip__Common::GenericParam(name, OPTIONAL<LibSip__Common::GenValue>(OMIT_VALUE));
      } else {
        LibSip__Common::GenValue g;
        g.tokenOrHost() = CHARSTRING(p->gvalue);
        l[i] = LibSip__Common::GenericParam(name, g);
      }
    } // End of 'for' statement
    p_list = OPTIONAL<LibSip__Common::AmpersandParam__List>(l);
  } else {
    p_list.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_ampersand_params: ", p_list);
} // End of method decode_ampersand_params

int sip_codec_headers::encode_option_tag_list(const LibSip__SIPTypesAndValues::OptionTag__List& p_options_tags, osip_message_t** p_sip_message) {
  loggers::get_instance().log(">>> sip_codec_headers::encode_option_tag_list");

  if (p_options_tags.size_of() != 0) {
    std::string str(static_cast<const char*>(*static_cast<const CHARSTRING*>(p_options_tags.get_at(0))));
    for (int i = 1; i < p_options_tags.size_of() - 1; i++) {
      str += "," + std::string(static_cast<const char*>(*static_cast<const CHARSTRING*>(p_options_tags.get_at(i))));
    } // End of 'for' statement
    osip_message_set_supported(*p_sip_message, str.c_str());
  } else {
    osip_message_set_supported(*p_sip_message, "");
  }
  return 0;
}

void sip_codec_headers::decode_option_tag_list(const char* p_list, OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>& p_options_tags)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_option_tag_list: %s", p_list);
  
  if (p_list != nullptr) {
    LibSip__SIPTypesAndValues::OptionTag__List l;
    std::string str(p_list);
    std::size_t i = str.find(","); // See RFC 3261 - Page 231 / LibSip__SIPTypesAndValues::OptionTag__List
    if (i == std::string::npos) { // Only on item
      l[0] = CHARSTRING(p_list);
    } else {
      int idx = 0;
      while(i != std::string::npos) {
        l[idx++] = CHARSTRING(str.substr(0, i - 1).c_str());
        str = str.substr(i + 1);
        loggers::get_instance().log("sip_codec_headers::decode_option_tag_list: New str: %s", str.c_str());
        i = str.find(",");
        loggers::get_instance().log("sip_codec_headers::decode_option_tag_list: New i: %d /%d", i, std::string::npos);
      } // End of 'while' statement
      l[idx] = CHARSTRING(str.c_str());
    }
    p_options_tags = OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>(l);
  } else {
    p_options_tags.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_option_tag_list: ", p_options_tags);
} // End of method decode_option_tag_list

void sip_codec_headers::decode_host_port(const char* p_host, const char* p_port, LibSip__SIPTypesAndValues::HostPort& p_host_port) {
  loggers::get_instance().log(">>> sip_codec_headers::decode_host_port");
  
  if (p_host == nullptr) {
    p_host_port.host().set_to_omit();
  } else {
    p_host_port.host() = OPTIONAL<CHARSTRING>(CHARSTRING(p_host));
  }
  OPTIONAL<INTEGER> port;
  if (p_port == nullptr) {
    p_host_port.portField().set_to_omit();
  } else {
    p_host_port.portField() = OPTIONAL<INTEGER>(INTEGER(std::stoi(p_port)));
  }

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_host_port: ", p_host_port);
} // End of method decode_host_port"

void sip_codec_headers::decode_uri(LibSip__SIPTypesAndValues::SipUrl& p_sip_url, const osip_uri_t* p_uri) {
  loggers::get_instance().log(">>> sip_codec_headers::decode_uri");
  
  // Scheme
  LibSip__SIPTypesAndValues::UriComponents uri_components;
  if (::osip_uri_get_scheme((osip_uri_t*)p_uri) != nullptr) {
    p_sip_url.scheme() = CHARSTRING(::osip_uri_get_scheme((osip_uri_t*)p_uri));
    loggers::get_instance().log_msg("sip_codec_headers::decode_uri: scheme: ", p_sip_url.scheme());
    // SipUriComponents
    if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "sip") == 0) {
      // User Info
      OPTIONAL<LibSip__SIPTypesAndValues::UserInfo> user_info;
      if (::osip_uri_get_username((osip_uri_t*)p_uri) == nullptr) {
        user_info.set_to_omit();
      } else {
        OPTIONAL<CHARSTRING> password;
        if (::osip_uri_get_password((osip_uri_t*)p_uri) == nullptr) {
          password.set_to_omit();
        } else {
          password = CHARSTRING(::osip_uri_get_password((osip_uri_t*)p_uri));
        }
        LibSip__SIPTypesAndValues::UserInfo u(CHARSTRING(::osip_uri_get_username((osip_uri_t*)p_uri)), password);
        user_info = OPTIONAL<LibSip__SIPTypesAndValues::UserInfo>(u);
      }
      loggers::get_instance().log_msg("sip_codec_headers::decode_uri: user_info: ", user_info);
      // HostPort
      LibSip__SIPTypesAndValues::HostPort host_port;
      decode_host_port(::osip_uri_get_host((osip_uri_t*)p_uri), ::osip_uri_get_port((osip_uri_t*)p_uri), host_port);
      loggers::get_instance().log_msg("sip_codec_headers::decode_uri: host_port: ", host_port);
      uri_components.sip() = LibSip__SIPTypesAndValues::SipUriComponents(user_info, host_port);
    } else if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "tel") == 0) {
      const LibSip__SIPTypesAndValues::TelUriComponents t;
      uri_components.tel() = t; // TODO To be done
      loggers::get_instance().error("sip_codec_headers::decode_uri: Unsupported LibSip__SIPTypesAndValues::UriComponents::ALT_tel");
    } else if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "urn") == 0) {
      char *buf = nullptr;
      if (::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buf) == 0) {
        std::string str(buf);
        loggers::get_instance().log("sip_codec_headers::decode_uri: str: %s", str.c_str());
        std::vector<std::string> output = converter::get_instance().split(str, ":");
        loggers::get_instance().log("sip_codec_headers::decode_uri: split size: %d", output.size());
        LibSip__SIPTypesAndValues::UrnUriComponents urn;
        urn.namespaceId() = output[1].c_str();
        std::string s(output[2]);
        for (size_t i = 3; i < output.size(); i++) {
          s += ":" + output[i];
        }
        urn.namespaceSpecificString() = s.c_str();
        uri_components.urn() = urn;
        osip_free(buf); // Macro
      } // TODO Check what to do in this case
    } else {
      char *buffer = nullptr;
      ::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buffer);
      uri_components.other() = CHARSTRING(buffer);
      osip_free(buffer); // Macro
    }
  } else {
    char *buffer = nullptr;
    ::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buffer);
    uri_components.other() = CHARSTRING(buffer);
    osip_free(buffer); // Macro
  }
  p_sip_url.components() = uri_components;
  // UrlParameters
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_uri->url_params, params);
  p_sip_url.urlParameters() = params;
  // Headers
  OPTIONAL<LibSip__Common::AmpersandParam__List> headers;
  decode_ampersand_params(p_uri->url_headers, headers);
  p_sip_url.headers() = headers;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_uri: ", p_sip_url);
} // End of method decode_uri

void sip_codec_headers::decode_accept_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Accept& p_accept_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_accept_header");

  // Sanity checks
  osip_accept_t *sip_accept = nullptr;
  ::osip_message_get_accept(p_sip_message, 0, &sip_accept);
  if (sip_accept == nullptr) {
    return;
  }
  
  // FieldName
  p_accept_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_E"));
  // AcceptArgs
  if (::osip_list_size(&p_sip_message->accepts) == 0) {
    p_accept_header.acceptArgs().set_to_omit();
  } else {
    LibSip__SIPTypesAndValues::AcceptBody__List bl;
    int x = 0;
    for (int i = 0; i < ::osip_list_size(&p_sip_message->accepts); i++) {
      const osip_accept_t* l = (const osip_accept_t*)::osip_list_get(&p_sip_message->accepts, i);
      if (l != nullptr) {
        loggers::get_instance().log("sip_codec_headers::decode_accept_header: %s=%s", l->type, l->subtype); // application=sdp,application/3gpp-ims+xml
        std::string str(l->subtype);
        std::size_t idx = str.find(",");
        if (idx == std::string::npos) {
          LibSip__SIPTypesAndValues::AcceptBody b;
          b.mediaRange() = CHARSTRING(l->type) + CHARSTRING("/") + CHARSTRING(l->subtype); // FIXME CHARSTRING(l->type || "=" ||  l->subtype)???
          if (osip_list_size(&(l->gen_params)) != 0) {
            // TODO
          } else {
            b.acceptParam().set_to_omit();
          }
          bl[x++] = b;
        } else {
          std::vector<std::string> output = converter::get_instance().split(str, ",");
          LibSip__SIPTypesAndValues::AcceptBody b;
          loggers::get_instance().log("sip_codec_headers::decode_accept_header: Processing %s=%s", l->type, output[0].c_str());
          b.mediaRange() = CHARSTRING(l->type) + CHARSTRING("/") + CHARSTRING(output[0].c_str()); // FIXME CHARSTRING(l->type || "=" ||  l->subtype)???
          if (osip_list_size(&(l->gen_params)) != 0) {
            // TODO
          } else {
            b.acceptParam().set_to_omit();
          }
          bl[x++] = b;
          for (unsigned int k = 1; k < output.size(); k++) {
            loggers::get_instance().log("sip_codec_headers::decode_accept_header: In loop, processing %s=%s", output[k].c_str());
            LibSip__SIPTypesAndValues::AcceptBody b;
            b.mediaRange() = CHARSTRING(output[k].c_str()); // FIXME CHARSTRING(l->type || "=" ||  l->subtype)???
            if (osip_list_size(&(l->gen_params)) != 0) {
              // TODO
            } else {
              b.acceptParam().set_to_omit();
            }
            bl[x++] = b;
          }
        }
      }
    } // End of 'for' statement
    p_accept_header.acceptArgs() = OPTIONAL<LibSip__SIPTypesAndValues::AcceptBody__List>(bl);
  }
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_accept_header: ", p_accept_header);
} // End of method decode_accept_header

void sip_codec_headers::decode_accept_contact_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptContact& p_accept_contact_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_accept_contact_header: %p", p_sip_message->contacts);

  // Sanity checks
  osip_header_t *sip_accept_contact = nullptr;
  osip_message_header_get_byname(p_sip_message,(const char *)"accept-contact", 0, &sip_accept_contact);
  if (sip_accept_contact == nullptr) {
    return;
  }
  
  loggers::get_instance().log("sip_codec_headers::decode_accept_contact_header: got it: %s:%s", sip_accept_contact->hname, sip_accept_contact->hvalue);
  if ((sip_accept_contact->hvalue == nullptr) || (strlen(sip_accept_contact->hvalue) == 0)) {
    loggers::get_instance().warning("sip_codec_headers::decode_accept_contact_header: Failed to decode Accept-Contact header");
    return;
  }
  // FieldName
  p_accept_contact_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_CONTACT_E"));
  // AcRcValue list
  std::string str(sip_accept_contact->hvalue);
  LibSip__SIPTypesAndValues::AcRcValue__List la;
  std::size_t idx = str.find(",");
  if (idx == std::string::npos) {
    // Split using ; as separator
    std::vector<std::string> output = converter::get_instance().split(str, ";");
    loggers::get_instance().log("sip_codec_headers::decode_accept_contact_header (1): Split size: %d", output.size());
    // The first parameter is the wild card
    LibSip__SIPTypesAndValues::AcRcValue acRcValue;
    acRcValue.wildcard() = CHARSTRING(output[0].c_str());
    if (output.size() > 1) {
      // Extract the other parameters
      LibSip__Common::SemicolonParam__List l;
      for (unsigned int i = 1; i < output.size(); i++) {
        LibSip__Common::GenericParam p;
        std::vector<std::string> o = converter::get_instance().split(output[i], "=");
        p.id() = CHARSTRING(o[0].c_str());
        if (o.size() == 1) { // No value
          p.paramValue().set_to_omit();
        } else {
          loggers::get_instance().log("sip_codec_headers::decode_accept_contact_header (1): o: %s:%s", o[0].c_str(), o[1].c_str());
          LibSip__Common::GenValue g;
          g.tokenOrHost() = CHARSTRING(o[1].c_str());
          p.paramValue() = OPTIONAL<LibSip__Common::GenValue>(g);
        }
        l[i - 1] = p;
      } // End of 'for' statement
      loggers::get_instance().log_msg("sip_codec_headers::decode_accept_contact_header: l: ", l);
      acRcValue.acRcParams() = OPTIONAL<LibSip__Common::SemicolonParam__List>(l);
    } else {
      acRcValue.acRcParams().set_to_omit();
    }
    la[0] = acRcValue;
  } else {
    // Split the list of values using , as separator
    std::vector<std::string> lop = converter::get_instance().split(str, ",");
    for (size_t s = 0; s < lop.size(); s++) {
      str = lop[s];
      // Extract parameters separated by ; into SemiColumn__List
      std::vector<std::string> output = converter::get_instance().split(str, ";");
      // The first parameter is the wild card
      LibSip__SIPTypesAndValues::AcRcValue acRcValue;
      acRcValue.wildcard() = CHARSTRING(output[0].c_str());
      if (output.size() > 1) {
        // Extract the other parameters
        LibSip__Common::SemicolonParam__List l;
        for (unsigned int i = 1; i < output.size(); i++) {
          LibSip__Common::GenericParam p;
          std::vector<std::string> o = converter::get_instance().split(output[i], "=");
          p.id() = CHARSTRING(o[0].c_str());
          if (o.size() == 1) { // No value
            p.paramValue().set_to_omit();
          } else {
            loggers::get_instance().log("sip_codec_headers::decode_accept_contact_header (2): o: %s:%s", o[0].c_str(), o[1].c_str());
            LibSip__Common::GenValue g;
            g.tokenOrHost() = CHARSTRING(o[1].c_str());
            p.paramValue() = OPTIONAL<LibSip__Common::GenValue>(g);
          }
          l[i - 1] = p;
        } // End of 'for' statement
        loggers::get_instance().log_msg("sip_codec_headers::decode_accept_contact_header: l: ", l);
        acRcValue.acRcParams() = OPTIONAL<LibSip__Common::SemicolonParam__List>(l);
      } else {
        acRcValue.acRcParams().set_to_omit();
      }
      la[s] = acRcValue;
    } // End of 'for'statement
  }
  p_accept_contact_header.acValues() = la;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_accept_contact_header: ", p_accept_contact_header);
} // End of method decode_accept_contact_header

void sip_codec_headers::decode_accept_encoding_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptEncoding& p_accept_encoding_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_accept_encoding_header: %p", p_sip_message->accept_encodings);

  // Sanity checks
  osip_accept_encoding_t *sip_accept_encoding = nullptr;
  ::osip_message_get_accept_encoding(p_sip_message, 0, &sip_accept_encoding);
  if (sip_accept_encoding == nullptr) {
    return;
  }
  // FieldName
  p_accept_encoding_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_ENCODING_E"));
  // ContentCoding
  p_accept_encoding_header.contentCoding().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_accept_encoding_header: ", p_accept_encoding_header);
} // End of method decode_accept_encoding_header

void sip_codec_headers::decode_accept_language_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptLanguage& p_accept_language_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_accept_language_header: %p", p_sip_message->accept_languages);

  // Sanity checks
  osip_accept_language_t *sip_accept_language = nullptr;
  ::osip_message_get_accept_language(p_sip_message, 0, &sip_accept_language);
  if (sip_accept_language == nullptr) {
    return;
  }
  
  // FieldName
  p_accept_language_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_LANGUAGE_E"));
  // LanguageBody
  p_accept_language_header.languageBody().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_accept_language_header: ", p_accept_language_header);
} // End of method decode_accept_language_header

void sip_codec_headers::decode_alert_info_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AlertInfo& p_alert_info_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_alert_info_header: %p", p_sip_message->alert_infos);

  // Sanity checks
  osip_alert_info_t *sip_alert_info = nullptr;
  ::osip_message_get_alert_info(p_sip_message, 0, &sip_alert_info);
  if (sip_alert_info == nullptr) {
    return;
  }
  
  // FieldName
  p_alert_info_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALERT_INFO_E"));
  // AlertInfoBody
  p_alert_info_header.alertInfoBody().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_alert_info_header: ", p_alert_info_header);
} // End of method decode_alert_info_header

void sip_codec_headers::decode_allow_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Allow& p_allow_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_allow_header: %p", p_sip_message->allows);

  // Sanity checks
  osip_allow_t *sip_allow = nullptr;
  ::osip_message_get_allow(p_sip_message, 0, &sip_allow);
  if (sip_allow == nullptr) {
    return;
  }
  
  // FieldName
  p_allow_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALLOW_E"));
  // Methods
  loggers::get_instance().log("sip_codec_headers::decode_allow_header: value: %s", sip_allow->value);
  if ((sip_allow->value == nullptr) || (strlen(sip_allow->value) == 0)) {
    p_allow_header.methods().set_to_omit();
  } else {
    LibSip__SIPTypesAndValues::Method__List m;
    std::string str(sip_allow->value);
    std::vector<std::string> output = converter::get_instance().split(str, ",");
    unsigned int i = 0;
    while (i < output.size()) {
      m[i] = CHARSTRING(output[i].c_str());
      i += 1;
    } // End of 'while' statement
    p_allow_header.methods() = OPTIONAL<LibSip__SIPTypesAndValues::Method__List>(m);
  }

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_allow_header: ", p_allow_header);
} // End of method decode_allow_header

void sip_codec_headers::decode_allow_events_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AllowEvents& p_allow_events_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_allow_events_header: %p", p_sip_message->headers);

  // Sanity checks
  osip_header_t *sip_allow_events = nullptr;
  osip_message_header_get_byname(p_sip_message,(const char *)"allow_events", 0, &sip_allow_events);
  if (sip_allow_events == nullptr) {
    return;
  }
  
  // FieldName
  p_allow_events_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALLOW_EVENTS_E"));
  // EventTypes
  LibSip__SIPTypesAndValues::EventType__List l;
  // TODO To be done
  l[0] = CHARSTRING("*");
  p_allow_events_header.eventTypes() = l;

  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_allow_events_header: ", p_allow_events_header);
} // End of method decode_allow_events_header

void sip_codec_headers::decode_authorization_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Authorization& p_authorization_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_authorization_header");

  // Sanity checks
  osip_authorization_t *sip_authorization = nullptr;
  ::osip_message_get_authorization(p_sip_message, 0, &sip_authorization);
  if (sip_authorization == nullptr) {
    return;
  }
  
  // FieldName
  p_authorization_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("AUTHORIZATION_E"));
  // CredentialsList
  LibSip__SIPTypesAndValues::CredentialsList l;
  int i = 0;
  do {
    loggers::get_instance().log("sip_codec_headers::decode_authorization_header: Processing item #%d", i);
    LibSip__SIPTypesAndValues::Credentials v;
    LibSip__Common::CommaParam__List cl;
    int j = 0;
    if (::osip_authorization_get_digest(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_digest(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("digest"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_username(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_username(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("username"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_uri(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_uri(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("uri"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_realm(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_realm(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("realm"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_nonce(sip_authorization) != nullptr) {
      loggers::get_instance().log("sip_codec_headers::decode_authorization_header: Decode nonce");
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_nonce(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("nonce"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_response(sip_authorization) != nullptr) {
      loggers::get_instance().log("sip_codec_headers::decode_authorization_header: Decode response");
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_response(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("response"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    loggers::get_instance().log_msg("sip_codec_headers::decode_authorization_header: New GenValue:", cl);
    v.digestResponse() = cl;
    loggers::get_instance().log_msg("sip_codec_headers::decode_authorization_header: New GenValue list:", cl);
    l[i++] = v;
    loggers::get_instance().log_msg("sip_codec_headers::decode_authorization_header: New CredentialsList:", l);
  } while(::osip_message_get_authorization(p_sip_message, i, &sip_authorization) == 0); // End of 'do-while' statement
  p_authorization_header.body() = l;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_authorization_header: ", p_authorization_header);
} // End of method decode_authorization_header

void sip_codec_headers::decode_call_id_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::CallId& p_call_id_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_call_id_header: %p", p_sip_message->call_id);
  
  // Sanity check
  if (p_sip_message->call_id == nullptr) {
    return;
  }

  // FieldName
  p_call_id_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CALL_ID_E"));
  // CallId
  char *buffer;
  ::osip_call_id_to_str(p_sip_message->call_id, &buffer);
  if (buffer != nullptr) {
    p_call_id_header.callid() = CHARSTRING(buffer);
    osip_free(buffer);
  }
} // End of method decode_call_id_header

void sip_codec_headers::decode_c_seq_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::CSeq& p_c_seq_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_c_seq_header");
  
  // Sanity check
  if (p_sip_message->cseq == nullptr) {
    return;
  }

  // FieldName
  p_c_seq_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CSEQ_E"));
  // Sequence number
  p_c_seq_header.seqNumber() = INTEGER(std::stoi(osip_cseq_get_number(p_sip_message->cseq)));
  // Method
  p_c_seq_header.method() = CHARSTRING(::osip_cseq_get_method(p_sip_message->cseq));
} // End of method decode_c_seq_header

void sip_codec_headers::decode_content_length_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::ContentLength& p_content_length_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_content_length_header");
  
  // Sanity check
  if (p_sip_message->content_length == nullptr) {
    return;
  }

  // FieldName
  p_content_length_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CONTENT_LENGTH_E"));
  // Length
  char *buffer;
  ::osip_content_length_to_str(p_sip_message->content_length, &buffer);
  if (buffer != nullptr) {
    p_content_length_header.len() = INTEGER(std::stoi(buffer));
    osip_free(buffer);
  }
} // End of method decode_content_length_header

void sip_codec_headers::decode_content_type_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::ContentType& p_content_type_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_content_type_header");
  
  // Sanity check
  if (p_sip_message->content_type == nullptr) {
    return;
  }

  // FieldName
  p_content_type_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CONTENT_TYPE_E"));
  // Type/SubType
  std::string str(p_sip_message->content_type->type);
  if (p_sip_message->content_type->subtype != nullptr) {
    str += "/";
    str += p_sip_message->content_type->subtype; 
    p_content_type_header.mTypeSubtype() = CHARSTRING(str.c_str());
  }
  // Parameters
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_sip_message->content_type->gen_params, params);
  p_content_type_header.mParams() = params;
} // End of method decode_content_type_header

void sip_codec_headers::decode_contact_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Contact& p_contact_header) {
  loggers::get_instance().log("sip_codec_headers::decode_contact_header");
  
  // Sanity check
  if (p_sip_message == nullptr) {
    return;
  }
  
  // FieldName
  p_contact_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CONTACT_E"));
  // ContactBody
  LibSip__SIPTypesAndValues::ContactAddress__List c;
  int i = 0;
  osip_contact_t* contact = nullptr;
  while (::osip_message_get_contact(p_sip_message, i, &contact) == 0) {
    LibSip__SIPTypesAndValues::ContactAddress contact_addr;
    LibSip__SIPTypesAndValues::SipUrl uri;
    decode_uri(uri, ::osip_contact_get_url((osip_contact_t*)contact));
    LibSip__SIPTypesAndValues::Addr__Union addr;
    OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
    if (::osip_contact_get_displayname((osip_contact_t*)contact) != nullptr) {
      LibSip__SIPTypesAndValues::DisplayName n;
      n.token() = CHARSTRING(::osip_contact_get_displayname((osip_contact_t*)contact));
      display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
      LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
      addr.nameAddr() = name_addr;
    } else {
      display_name.set_to_omit();
      addr.addrSpecUnion() = uri;
    }
    contact_addr.addressField() = addr;
    // Params
    OPTIONAL<LibSip__Common::SemicolonParam__List> params;
    decode_semi_colon_params(contact->gen_params, params);
    contact_addr.contactParams() = params;
    
    c[i++] = contact_addr;
  } // End of 'while' statement
  p_contact_header.contactBody().contactAddresses() = c;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_contact_header: ", p_contact_header);
} // End of method decode_contact_header

void sip_codec_headers::decode_from_header(const osip_from_t* p_sip_from, LibSip__SIPTypesAndValues::From& p_from_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_from_header");
  
  // Sanity check
  if (p_sip_from == nullptr) {
    return;
  }

  // FieldName
  p_from_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("FROM_E"));
  // Addr_Union
  LibSip__SIPTypesAndValues::SipUrl uri;
  decode_uri(uri, ::osip_from_get_url((osip_from_t*)p_sip_from));
  LibSip__SIPTypesAndValues::Addr__Union addr;
  OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
  if (::osip_from_get_displayname((osip_from_t*)p_sip_from) != nullptr) {
    LibSip__SIPTypesAndValues::DisplayName n;
    n.token() = CHARSTRING(::osip_from_get_displayname((osip_from_t*)p_sip_from));
    display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
    LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
    addr.nameAddr() = name_addr;
  } else {
    display_name.set_to_omit();
    addr.addrSpecUnion() = uri;
  }
  p_from_header.addressField() = addr;
  // Params
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_sip_from->gen_params, params);
  p_from_header.fromParams() = params;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_from_header: ", p_from_header);
} // End of method decode_from_header

void sip_codec_headers::decode_max_forwards_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::MaxForwards& p_max_forwards_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_max_forwards_header");

  // Sanity checks
  osip_header_t *dest = nullptr;
  osip_message_get_max_forwards(p_sip_message, 0, &dest);
  if (dest == nullptr) {
    return;
  }
  
  // FieldName
  p_max_forwards_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("MAX_FORWARDS_E"));
  // Forwards
  loggers::get_instance().log("sip_codec_headers::decode_max_forwards_header: got it: %s:%s", dest->hname, dest->hvalue);
  p_max_forwards_header.forwards() = INTEGER(std::stoi(dest->hvalue));
} // End of method decode_max_forwards_header

void sip_codec_headers::decode_min_se_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::MinSE& p_min_se_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_min_se_header");

  // Sanity checks
  osip_header_t *dest = nullptr;
  osip_message_header_get_byname((osip_message_t*)p_sip_message,(const char *)"min-se", 0, &dest); // TODO Create osip_message_[g|s]et_min_se
  if (dest == nullptr) {
    return;
  }
  
  // FieldName
  p_min_se_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("MIN_SE_E"));
  
  std::string fisrt_parameter;
  LibSip__Common::SemicolonParam__List others_params;
  decode_semi_colon_params(dest->hvalue, fisrt_parameter, others_params);
  p_min_se_header.deltaSec() = CHARSTRING(fisrt_parameter.c_str());
  if (others_params.is_bound()) {
    p_min_se_header.minSeParam() = OPTIONAL<LibSip__Common::SemicolonParam__List>(others_params);
  } else {
    p_min_se_header.minSeParam().set_to_omit();
  }
} // End of method decode_min_se_header

void sip_codec_headers::decode_route_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Route& p_route_header) {
  loggers::get_instance().log("sip_codec_headers::decode_route_header");
  // Sanity check
  if (p_sip_message == nullptr) {
    return;
  }
  
  // FieldName
  p_route_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ROUTE_E"));
  // RouteBody
  LibSip__SIPTypesAndValues::RouteBody__List c;
  int i = 0;
  osip_route_t* route = nullptr;
  while (::osip_message_get_route(p_sip_message, i, &route) == 0) {
    LibSip__SIPTypesAndValues::RouteBody route_body;
    // Decode uri
    LibSip__SIPTypesAndValues::SipUrl uri;
    decode_uri(uri, ::osip_from_get_url((osip_from_t*)route));
    // Decode display name
    OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
    if (::osip_from_get_displayname((osip_route_t*)route) != nullptr) {
      LibSip__SIPTypesAndValues::DisplayName n;
      n.token() = CHARSTRING(::osip_from_get_displayname((osip_route_t*)route));
      display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
    } else {
      display_name.set_to_omit();
    }
    route_body.nameAddr() = LibSip__SIPTypesAndValues::NameAddr(display_name, uri);
    // Params
    OPTIONAL<LibSip__Common::SemicolonParam__List> params;
    decode_semi_colon_params(route->gen_params, params);
    route_body.rrParam() = params;
    
    c[i++] = route_body;
  } // End of 'while' statement
  p_route_header.routeBody() = c;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_route_header: ", p_route_header);
} // End of method decode_route_header

void sip_codec_headers::decode_session_expires_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::SessionExpires& p_session_expires_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_session_expires_header");

  // Sanity checks
  osip_header_t *dest = nullptr;
  osip_message_header_get_byname((osip_message_t*)p_sip_message,(const char *)"session-expires", 0, &dest); // TODO Create osip_message_[g|s]et_session_expires
  if (dest == nullptr) {
    return;
  }
  
  // FieldName
  p_session_expires_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("SESSION_EXPIRES_E"));
  
  std::string fisrt_parameter;
  LibSip__Common::SemicolonParam__List others_params;
  decode_semi_colon_params(dest->hvalue, fisrt_parameter, others_params);
  p_session_expires_header.deltaSec() = CHARSTRING(fisrt_parameter.c_str()); 
  if (others_params.is_bound()) {
    p_session_expires_header.seParam() = OPTIONAL<LibSip__Common::SemicolonParam__List>(others_params); 
  } else {
    p_session_expires_header.seParam().set_to_omit();
  }
  
  // loggers::get_instance().log("sip_codec_headers::decode_session_expires_header: got it: %s:%s", dest->hname, dest->hvalue);
  // std::size_t idx = str.find(";");
  // if (idx == std::string::npos) {
  //   p_session_expires_header.deltaSec() = dest->hvalue;
  //   p_session_expires_header.seParam().set_to_omit();
  // } else {
  //   std::vector<std::string> output = converter::get_instance().split(str, ";");
  //   p_session_expires_header.deltaSec() = CHARSTRING(output[0].c_str());
  //   LibSip__Common::SemicolonParam__List l;
  //   for (unsigned int i = 1; i < output.size(); i++) {
  //     std::vector<std::string> o = converter::get_instance().split(output[i], "=");
  //     LibSip__Common::GenValue g;
  //     g.tokenOrHost() = CHARSTRING(o[1].c_str());
  //     l[i - 1] = LibSip__Common::GenericParam(CHARSTRING(o[0].c_str()), g);
  //   } // End of 'for' statement
  //   p_session_expires_header.seParam() = OPTIONAL<LibSip__Common::SemicolonParam__List>(l);
  // }
} // End of method decode_session_expires_header

void sip_codec_headers::decode_supported_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Supported& p_supported_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_supported_header");

  // Sanity checks
  osip_header_t *dest = nullptr;
  osip_message_get_supported(p_sip_message, 0, &dest);
  if (dest == nullptr) {
    return;
  }
  
  // FieldName
  p_supported_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("SUPPORTED_E"));
  // OptionTags
  loggers::get_instance().log("sip_codec_headers::decode_supported_header: got it: %s:%s", dest->hname, dest->hvalue);
  OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>& l = p_supported_header.optionsTags();
  decode_option_tag_list(dest->hvalue, l);
} // End of method decode_supported_header

void sip_codec_headers::decode_to_header(const osip_to_t* p_sip_to, LibSip__SIPTypesAndValues::To& p_to_header)
{
  loggers::get_instance().log("sip_codec_headers::decode_to_header");
  
  // Sanity check
  if (p_sip_to == nullptr) {
    return;
  }

  // FieldName
  p_to_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("TO_E"));
  // Addr_Union
  LibSip__SIPTypesAndValues::SipUrl uri;
  decode_uri(uri, ::osip_to_get_url((osip_to_t*)p_sip_to));
  LibSip__SIPTypesAndValues::Addr__Union addr;
  OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
  if (::osip_to_get_displayname((osip_to_t*)p_sip_to) != nullptr) {
    LibSip__SIPTypesAndValues::DisplayName n;
    n.token() = CHARSTRING(::osip_to_get_displayname((osip_to_t*)p_sip_to));
    display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
    LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
    addr.nameAddr() = name_addr;
  } else {
    display_name.set_to_omit();
    addr.addrSpecUnion() = uri;
  }
  p_to_header.addressField() = addr;
  // Params
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_sip_to->gen_params, params);
  p_to_header.toParams() = params;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_to_header: ", p_to_header);
} // End of method decode_to_header

void sip_codec_headers::decode_via_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Via& p_via_header)
{
  loggers::get_instance().log(">>> sip_codec_headers::decode_via_header: %p", p_sip_message->vias);

  // Sanity checks
  osip_via_t *sip_via = nullptr;
  ::osip_message_get_via(p_sip_message, 0, &sip_via);
  if (sip_via == nullptr) {
    return;
  }
  
  // FieldName
  p_via_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("VIA_E"));
  // ViaBody
  LibSip__SIPTypesAndValues::ViaBody__List l;
  int i = 0;
  do {
    loggers::get_instance().log("sip_codec_headers::decode_via_header: Processing item #%d", i);
    LibSip__SIPTypesAndValues::ViaBody v;
    v.sentProtocol() = LibSip__SIPTypesAndValues::SentProtocol(
                                                               CHARSTRING(::via_get_protocol(sip_via)),
                                                               CHARSTRING(::via_get_version(sip_via)),
                                                               CHARSTRING(::via_get_comment(sip_via))
                                                               );
    OPTIONAL<LibSip__SIPTypesAndValues::HostPort> host_port;
    decode_host_port(::via_get_host(sip_via), ::via_get_port(sip_via), host_port);
    loggers::get_instance().log_msg("sip_codec_headers::decode_via_uri: host_port: ", host_port);
    v.sentBy() = host_port;
    OPTIONAL<LibSip__Common::SemicolonParam__List> params;
    decode_semi_colon_params(sip_via->via_params, params);
    v.viaParams() = params;
    l[i++] = v;
  } while(::osip_message_get_via(p_sip_message, i, &sip_via) == 0); // End of 'while' statement
  p_via_header.viaBody() = l;
  
  loggers::get_instance().log_msg("<<< sip_codec_headers::decode_via_header: ", p_via_header);
} // End of method decode_via_header
