#include "LibSip_SIPTypesAndValues.hh"
#include "LibSip_Common.hh"

#include "sip_codec_response.hh"
#include "loggers.hh"

#include "converter.hh"

int sip_codec_response::encode (const LibSip__SIPTypesAndValues::Response& msg, OCTETSTRING& data)
{
  loggers::get_instance().log_msg(">>> sip_codec_response::encode: ", (const Base_Type&)msg);

  osip_message_t* sip_message = nullptr; // TODO Use smart pointer
  ::parser_init();
  int result = ::osip_message_init(&sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_response::encode: Failed to initialise internal data structures");
    return -1;
  }

  if (encode_response(msg, sip_message) == -1) {
    loggers::get_instance().warning("sip_codec_response::encode: Failed to encode Response");
    ::osip_message_free(sip_message);
    sip_message = nullptr;
    return -1;
  }
  
  char* buffer = nullptr;
  size_t length = 0;
  result = ::osip_message_to_str(sip_message, &buffer, &length);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_response::encode: Failed to encode data structures");
    ::osip_message_free(sip_message);
    return -1;
  }
  loggers::get_instance().log("sip_codec_response::encode: Message:%s", buffer);
  data = OCTETSTRING(length, (const unsigned char*)buffer);
  osip_free(buffer);
  ::osip_message_free(sip_message);
  
  loggers::get_instance().log_msg("<<< sip_codec_response::encode: data=", data);
  return result;
}

int sip_codec_response::decode (const OCTETSTRING& data, LibSip__SIPTypesAndValues::Response& msg, params* p_params)
{
  loggers::get_instance().log_msg(">>> sip_codec_response::decode: data=", data);

  // Sanity checks
  if (data.lengthof() == 0) {
    loggers::get_instance().warning("sip_codec_response::decode: Wrong parameters");
    return -1;
  }

  osip_message_t* sip_message;
  ::parser_init();
  int result = ::osip_message_init(&sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_response::decode: Failed to initialise internal data structures");
    return -1;
  }
  result = ::osip_message_parse(sip_message, (const char*)static_cast<const unsigned char*>(data), data.lengthof());
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_response::decode: Failed to parse SIP message");
    ::osip_message_free(sip_message);
    return -1;
  }

  // Fill StatusLine
  decode_status_line(sip_message, msg);

  // Fill Headers
  decode_response_headers(sip_message, msg);

  // Fill MessageBody
  decode_message_body(sip_message, msg.messageBody());
  
  // Fill Payload
  decode_payload(sip_message, msg);
  
  ::osip_message_free(sip_message);
  loggers::get_instance().log_msg("sip_codec_response::decode: ", msg);
  
  return 0;
}

int sip_codec_response::encode_response(const LibSip__SIPTypesAndValues::Response& p_response, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_response::encode_response");
  
  encode_status_line(p_response.statusLine(), p_sip_message);
  
  encode_response_headers(p_response.msgHeader(), p_sip_message);
  
  if (p_response.messageBody().is_present()) {
    const LibSip__MessageBodyTypes::MessageBody& m = static_cast<const LibSip__MessageBodyTypes::MessageBody&>(*p_response.messageBody().get_opt_value());
    encode_message_body(m, p_sip_message);
  }
  
  if (p_response.payload().is_present()) {
    const LibSip__SIPTypesAndValues::Payload& p = static_cast<const LibSip__SIPTypesAndValues::Payload&>(*p_response.payload().get_opt_value());
    encode_response_payload(p, p_sip_message);
  }

  return 0;
} // End of method encode_response

int sip_codec_response::encode_status_line(const LibSip__SIPTypesAndValues::StatusLine& p_status_line, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_response::encode_status_line");

  // Sip version
  char* p = (char*)osip_malloc(p_status_line.sipVersion().lengthof() + 1);
  ::strcpy(p, (char*)static_cast<const char*>(p_status_line.sipVersion()));
  ::osip_message_set_version(p_sip_message, p);

  // Status code
  ::osip_message_set_status_code(p_sip_message, static_cast<const int>(p_status_line.statusCode()));
  
  // Reason phrase
  p = (char*)osip_malloc(p_status_line.reasonPhrase().lengthof() + 1);
  ::strcpy(p, (char*)static_cast<const char*>(p_status_line.reasonPhrase()));
  ::osip_message_set_version(p_sip_message, p);

  return 0;
} // End of method encode_status_line

int sip_codec_response::encode_response_headers(const LibSip__SIPTypesAndValues::MessageHeader& p_msg_header, osip_message_t* p_sip_message)
{
  loggers::get_instance().log_msg(">>> sip_codec_response::encode_response_headers: ", p_msg_header);
  
  // Encode mandatory fields
  // From
  osip_from_t* from_header = nullptr;
  if (encode_from_header(p_msg_header.fromField(), &from_header) == -1) {
    loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode From header");
    return -1;
  }
  char* hvalue;
  ::osip_from_to_str(from_header, &hvalue);
  loggers::get_instance().log("sip_codec_response::encode_response_headers: From:%s", hvalue);
  ::osip_message_set_from(p_sip_message, hvalue);
  ::osip_from_free(from_header);
  osip_free(hvalue);
  
  // To
  osip_to_t* to_header = nullptr;
  if (encode_to_header(p_msg_header.toField(), &to_header) == -1) {
    loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode To header");
    return -1;
  }
  int r = ::osip_to_to_str(to_header, &hvalue);
  loggers::get_instance().log("sip_codec_response::encode_response_headers: To:'%s'- result:%d", hvalue, r);
  ::osip_message_set_to(p_sip_message, hvalue);
  ::osip_to_free(to_header);
  osip_free(hvalue);

  // Via
  osip_via_t* via_header = nullptr;
  if (encode_via_header(p_msg_header.via(), &via_header) == -1) {
    loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode Via header");
    return -1;
  }
  ::osip_via_to_str(via_header, &hvalue);
  loggers::get_instance().log("sip_codec_response::encode_response_headers: Via:%s", hvalue);
  ::osip_message_set_via(p_sip_message, hvalue);
  ::osip_via_free(via_header);
  osip_free(hvalue);

  // Encode Optional fields
  loggers::get_instance().log("sip_codec_response::encode_response_headers: Encode Optional fields");
  // Authorization
  if (p_msg_header.authorization().is_present()) {
    osip_authorization_t* authorization_header = nullptr;
    if (encode_authorization_header(p_msg_header.authorization(), &authorization_header) == -1) {
      loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode Authorization header");
      return -1;
    }
    int result = ::osip_authorization_to_str(authorization_header, &hvalue);
    loggers::get_instance().log("sip_codec_response::encode_response_headers: Authorization: %s - %d", hvalue, result);
    result = ::osip_message_set_authorization(p_sip_message, hvalue);
    loggers::get_instance().log("sip_codec_response::encode_response_headers: SIP Authorization: %p - %d", p_sip_message->authorizations, result);
    ::osip_authorization_free(authorization_header);
    osip_free(hvalue);
  }
  
  // CallId
  if (p_msg_header.callId().is_present()) {
    osip_call_id_t* call_id_header;
    if (encode_call_id_header(p_msg_header.callId(), &call_id_header) == -1) {
      loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode Call_Id header");
      return -1;
    }
    ::osip_call_id_to_str(call_id_header, &hvalue);
    loggers::get_instance().log("sip_codec_response::encode_response_headers: Call_Id:%s", hvalue);
    int result = ::osip_message_set_call_id(p_sip_message, hvalue);
    loggers::get_instance().log("sip_codec_response::encode_response_headers: SIP Call_Id: %p - %d", p_sip_message->call_id, result);
    ::osip_call_id_free(call_id_header);
    osip_free(hvalue);
  }
  
  // ContactHeader
  if (p_msg_header.contact().is_present()) {
    osip_contact_t* contact_header;
    if (encode_contact_header(p_msg_header.contact(), &contact_header) == -1) {
      loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode Contact header");
      return -1;
    }
    ::osip_contact_to_str(contact_header, &hvalue);
    loggers::get_instance().log("sip_codec_response::encode_response_headers: Contact: %s", hvalue);
    int result = ::osip_message_set_contact(p_sip_message, hvalue);
    loggers::get_instance().log("sip_codec_response::encode_response_headers: SIP Contact: %p - %d", p_sip_message->contacts, result);
    ::osip_contact_free(contact_header);
    osip_free(hvalue);
  }
  
  // CSeq
  osip_cseq_t* cseq_header = nullptr;
  if (encode_c_seq_header(p_msg_header.cSeq(), &cseq_header) == -1) {
    loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode CSeq header");
    return -1;
  }
  ::osip_cseq_to_str(cseq_header, &hvalue);
  loggers::get_instance().log("sip_codec_response::encode_response_headers: Cseq:%s", hvalue);
  ::osip_message_set_cseq(p_sip_message, hvalue);
  ::osip_cseq_free(cseq_header);
  osip_free(hvalue);

  // MaxForwards
  if (p_msg_header.maxForwards().is_present()) {
    if (encode_max_forwards_header(p_msg_header.maxForwards(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode MaxForwards header");
      return -1;
    }
  }
  
  // Supported
  if (p_msg_header.supported().is_present()) {
    if (encode_supported_header(p_msg_header.supported(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode Supported header");
      return -1;
    }
  }
  
  // PAccessNetworkInfo
  if (p_msg_header.pAccessNetworkInfo().is_present()) {
    if (encode_p_access_network_info_header(p_msg_header.pAccessNetworkInfo(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_response::encode_response_headers: Faile to encode PAccessNetworkInfo header");
      return -1;
    }
  }
  
  // TODO continue
  const osip_list_t* p = &(p_sip_message->headers);
  unsigned int pos = 0;
  unsigned int size = ::osip_list_size(p);
  loggers::get_instance().log("Unknown header size: %d\n", size);
  while (pos < size) {
    const osip_header_t* header = (const osip_header_t*)osip_list_get(p, pos++);
    loggers::get_instance().log("%p: hname=%s : hvalue=%s\n", header, header->hname, header->hvalue);
  } // End of 'while' statement
  
  return 0;
} // End of method encode_response_headers

int sip_codec_response::encode_response_payload(const LibSip__SIPTypesAndValues::Payload& p_payload, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_response::encode_response_payload");

  return 0;
} // End of method encode_response_payload

void sip_codec_response::decode_status_line(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Response& p_response)
{
  loggers::get_instance().log(">>> sip_codec_response::decode_status_line");
  
  LibSip__SIPTypesAndValues::StatusLine status_line;
  status_line.sipVersion() = CHARSTRING(::osip_message_get_version(p_sip_message));
  loggers::get_instance().log_msg("sip_codec_response::decode_status_line: sipVersion: ", status_line.sipVersion());
  status_line.statusCode() = INTEGER(osip_message_get_status_code(p_sip_message));
  loggers::get_instance().log_msg("sip_codec_response::decode_status_line: statusCode: ", status_line.statusCode());
  status_line.reasonPhrase() = CHARSTRING(::osip_message_get_reason_phrase(p_sip_message));
  loggers::get_instance().log_msg("sip_codec_response::decode_status_line: reasonPhrase: ", status_line.reasonPhrase());
  p_response.statusLine() = status_line;

  loggers::get_instance().log_msg("<<< sip_codec_response::decode_status_line: ", p_response.statusLine());
} // End of method decode_status_line

void sip_codec_response::decode_response_headers(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Response& p_response)
{
  loggers::get_instance().log(">>> sip_codec_response::decode_response_headers");
  
  LibSip__SIPTypesAndValues::MessageHeader headers;
  // Decode mandatory fields
  // From
  LibSip__SIPTypesAndValues::From from_header;
  decode_from_header(::osip_message_get_from(p_sip_message), from_header);
  if (from_header.is_value()) {
    headers.fromField() = from_header;
  } else {
    headers.fromField().set_to_omit();
  }
  // To
  LibSip__SIPTypesAndValues::To to_header;
  decode_to_header(::osip_message_get_to(p_sip_message), to_header);
  if (to_header.is_value()) {
    headers.toField() = to_header;
  } else {
    headers.toField().set_to_omit();
  }
  // Via
  LibSip__SIPTypesAndValues::Via via_header;
  decode_via_header(p_sip_message, via_header);
  if (via_header.is_value()) {
    headers.via() = via_header;
  } else {
    headers.via().set_to_omit();
  }

  // Decode Optional fields
  LibSip__SIPTypesAndValues::Accept accept_header;
  decode_accept_header(p_sip_message, accept_header);
  if (accept_header.is_value()) {
    headers.accept() = accept_header;
  } else {
    headers.accept().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptContact accept_contact_header;
  decode_accept_contact_header(p_sip_message, accept_contact_header);
  if (accept_contact_header.is_value()) {
    headers.acceptContact() = accept_contact_header;
  } else {
    headers.acceptContact().set_to_omit();
  }
  
  LibSip__SIPTypesAndValues::AcceptEncoding accept_encoding_header;
  decode_accept_encoding_header(p_sip_message, accept_encoding_header);
  if (accept_encoding_header.is_value()) {
    headers.acceptEncoding() = accept_encoding_header;
  } else {
    headers.acceptEncoding().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptLanguage accept_language_header;
  decode_accept_language_header(p_sip_message, accept_language_header);
  if (accept_language_header.is_value()) {
    headers.acceptLanguage() = accept_language_header;
  } else {
    headers.acceptLanguage().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AlertInfo alert_info_header;
  decode_alert_info_header(p_sip_message, alert_info_header);
  if (alert_info_header.is_value()) {
    headers.alertInfo() = alert_info_header;
  } else {
    headers.alertInfo().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Allow allow_header;
  decode_allow_header(p_sip_message, allow_header);
  if (allow_header.is_value()) {
    headers.allow() = allow_header;
  } else {
    headers.allow().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AllowEvents allow_events_header;
  decode_allow_events_header(p_sip_message, allow_events_header);
  if (allow_events_header.is_value()) {
    headers.allowEvents() = allow_events_header;
  } else {
    headers.allowEvents().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Authorization authorization_header;
  decode_authorization_header(p_sip_message, authorization_header);
  if (authorization_header.is_value()) {
    headers.authorization() = authorization_header;
  } else {
    headers.authorization().set_to_omit();
  }
  
  headers.authenticationInfo().set_to_omit();

  LibSip__SIPTypesAndValues::CallId call_id_header;
  decode_call_id_header(p_sip_message, call_id_header);
  if (call_id_header.is_value()) {
    headers.callId() = call_id_header;
  } else {
    headers.callId().set_to_omit();
  }
  
  headers.callInfo().set_to_omit();

  LibSip__SIPTypesAndValues::Contact contact_header;
  decode_contact_header(p_sip_message, contact_header);
  if (contact_header.is_value()) {
    headers.contact() = contact_header;
  } else {
    headers.contact().set_to_omit();
  }
  
  headers.contentDisposition().set_to_omit();
  headers.contentEncoding().set_to_omit();
  headers.contentLanguage().set_to_omit();
  
  LibSip__SIPTypesAndValues::ContentLength content_length_header;
  decode_content_length_header(p_sip_message, content_length_header);
  if (content_length_header.is_value()) {
    headers.contentLength() = content_length_header;
  } else {
    headers.contentLength().set_to_omit();
  }

  LibSip__SIPTypesAndValues::ContentType content_type_header;
  decode_content_type_header(p_sip_message, content_type_header);
  if (content_type_header.is_value()) {
    headers.contentType() = content_type_header;
  } else {
    headers.contentType().set_to_omit();
  }

  LibSip__SIPTypesAndValues::CSeq c_seq_header;
  decode_c_seq_header(p_sip_message, c_seq_header);
  if (c_seq_header.is_value()) {
    headers.cSeq() = c_seq_header;
  } else {
    headers.cSeq().set_to_omit();
  }
  
  headers.date().set_to_omit();
  headers.errorInfo().set_to_omit();
  headers.event().set_to_omit();
  headers.expires().set_to_omit();
  headers.featureCaps().set_to_omit();
  headers.geolocation().set_to_omit();
  headers.geolocationRouting().set_to_omit();
  headers.historyInfo().set_to_omit();
  headers.infoPackage().set_to_omit();
  headers.inReplyTo().set_to_omit();

  LibSip__SIPTypesAndValues::MaxForwards max_forwards_header;
  decode_max_forwards_header(p_sip_message, max_forwards_header);
  if (max_forwards_header.is_value()) {
    headers.maxForwards() = max_forwards_header;
  } else {
    headers.maxForwards().set_to_omit();
  }

  headers.mimeVersion().set_to_omit();
  headers.minExpires().set_to_omit();
  headers.minSE().set_to_omit();
  headers.organization().set_to_omit();


  // TODO
  headers.pAccessNetworkInfo().set_to_omit();


  
  headers.pAssertedID().set_to_omit();
  headers.pAssertedService().set_to_omit();
  headers.pAssociatedURI().set_to_omit();
  headers.path().set_to_omit();
  headers.pCalledPartyID().set_to_omit();
  headers.pChargingFunctionAddresses().set_to_omit();
  headers.pChargingVector().set_to_omit();
  headers.pEarlyMedia().set_to_omit();
  headers.pMediaAuthorization().set_to_omit();
  headers.pPreferredID().set_to_omit();
  headers.pPreferredService().set_to_omit();
  headers.priority().set_to_omit();
  headers.privacy().set_to_omit();
  headers.proxyAuthenticate().set_to_omit();
  headers.proxyAuthorization().set_to_omit();
  headers.proxyRequire().set_to_omit();
  headers.pVisitedNetworkID().set_to_omit();
  headers.rAck().set_to_omit();
  headers.rSeq().set_to_omit();
  headers.reason().set_to_omit();
  headers.recordRoute().set_to_omit();
  headers.recvInfo().set_to_omit();
  headers.requestDisposition().set_to_omit();
  headers.referredBy().set_to_omit();
  headers.referTo().set_to_omit();
  headers.referSub().set_to_omit();
  headers.replaces().set_to_omit();
  headers.replyTo().set_to_omit();
  headers.require().set_to_omit();
  headers.retryAfter().set_to_omit();
  headers.route().set_to_omit();
  headers.securityClient().set_to_omit();
  headers.securityServer().set_to_omit();
  headers.securityVerify().set_to_omit();
  headers.server().set_to_omit();
  headers.serviceRoute().set_to_omit();
  headers.sessionExpires().set_to_omit();
  headers.sessionId().set_to_omit();
  headers.sipETag().set_to_omit();
  headers.sipIfMatch().set_to_omit();
  headers.subject().set_to_omit();
  headers.subscriptionState().set_to_omit();
  
  LibSip__SIPTypesAndValues::Supported supported_header;
  decode_supported_header(p_sip_message, supported_header);
  if (supported_header.is_value()) {
    headers.supported() = supported_header;
  } else {
    headers.supported().set_to_omit();
  }
  
  headers.timestamp__().set_to_omit();
  headers.unsupported().set_to_omit();
  headers.userToUser().set_to_omit();
  headers.userAgent().set_to_omit();
  headers.warning().set_to_omit();
  headers.wwwAuthenticate().set_to_omit();
  headers.resourcePriority().set_to_omit();
  headers.answerMode().set_to_omit();
  headers.privAnswerMode().set_to_omit();
  headers.targetDialog().set_to_omit();
  headers.pAnswerState().set_to_omit();
  headers.undefinedHeader__List().set_to_omit();
  
  p_response.msgHeader() = headers;

  // List unprocessed headers
  int i = 0;
  osip_header_t* h_;
  while (::osip_message_get_header(p_sip_message, i++, &h_) == 0) {
    loggers::get_instance().warning("sip_codec_response::decode_response_headers: '%s'/'%s' was not processed", h_->hname, h_->hvalue);
  } // End of 'while' statement
  
  loggers::get_instance().log_msg("<<< sip_codec_response::decode_response_headers: ", p_response);
} // End of method decode_response_headers

void sip_codec_response::decode_payload(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Response& p_response)
{
  loggers::get_instance().log(">>> sip_codec_response::decode_payload");

  p_response.payload().set_to_omit();
  
} // End of method decode_payload
